import os
import sys
import json
import logging
import base64
import datetime
import time
from io import StringIO
from urllib.parse import urlencode

from zevon.sample_event import sample
from flask import Flask


from werkzeug.wrappers import Request

OCTET_STREAM = 'application/octet-stream'

binary_things = [
    'image',
    'font'
]

init_time = int(time.time())
logger = logging.getLogger(__name__)

log_levels = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR
}
log_level = os.environ.get('ZEVON_LOG_LEVEL', 'WARNING')
logger.setLevel(log_levels.get(log_level, logging.WARNING))


def json_converter(o):
    '''
    Helper thing to convert dates for JSON modulet.

    Args:
        o - the thing to dump as string.

    Returns:
        if an instance of datetime the a string else None
    '''
    if isinstance(o, datetime.datetime):
        return o.__str__()
    elif isinstance(o, StringIO):
        return o.getvalue()

    return None


def create_request_data(event):
    request_data = {}
    try:
        request_data['id'] = event.get('requestContext', {}).get('requestId', None)
        request_data['requestTimeEpoch'] = event.get('requestContext', {}).get('requestTimeEpoch', None)
        request_data['initTimeEpoch'] = init_time
        request_data['lambdaAge'] = int(time.time()) - init_time
        request_data['sourceIp'] = event.get('requestContext', {}).get('identity', {}).get('sourceIp', None)
        request_data['userArn'] = event.get('requestContext', {}).get('identity', {}).get('userArn', None)
        request_data['userAgent'] = event.get('requestContext', {}).get('identity', {}).get('userAgent', None)
    except Exception as ruh_roh_shaggy:
        logger.error(ruh_roh_shaggy, exc_info=False)

    return request_data


def make_environ(event):
    logger.info(json.dumps(event, indent=2))
    environ = {}

    caller = event.get('requestContext', {}).get('identity', {}).get('userArn', 'anonymous')

    if caller is None:
        caller = 'anonymous'

    event['headers']['x-zvn-caller'] = caller
    logger.info(f'{caller=} [e00da0624721]')
    for hdr_name, hdr_value in event['headers'].items():
        hdr_name = hdr_name.replace('-', '_').upper()
        if hdr_name in ['CONTENT_TYPE', 'CONTENT_LENGTH']:
            environ[hdr_name] = hdr_value
            continue

        http_hdr_name = 'HTTP_{}'.format(hdr_name)
        environ[http_hdr_name] = hdr_value

    is_encoded = event.get('isBase64Encoded', False)

    wrk = event.get('body', '')
    if is_encoded and wrk:
        event_body = base64.b64decode(wrk).decode()
    else:
        event_body = wrk

    environ['REQUEST_METHOD'] = event['httpMethod']
    environ['PATH_INFO'] = event['path']
    environ['REMOTE_ADDR'] = event['requestContext']['identity']['sourceIp']
    environ['HOST'] = '%(HTTP_HOST)s:%(HTTP_X_FORWARDED_PORT)s' % environ
    environ['SCRIPT_NAME'] = ''
    environ['SERVER_PORT'] = environ['HTTP_X_FORWARDED_PORT']
    environ['SERVER_PROTOCOL'] = 'HTTP/1.1'
    environ['wsgi.url_scheme'] = environ['HTTP_X_FORWARDED_PROTO']
    environ['wsgi.input'] = StringIO(event_body)
    environ['wsgi.version'] = (1, 0)
    environ['wsgi.errors'] = sys.stderr
    environ['wsgi.multithread'] = False
    environ['wsgi.run_once'] = True
    environ['wsgi.multiprocess'] = False

    query_string = event['queryStringParameters']
    if query_string:
        environ['QUERY_STRING'] = urlencode(query_string)
    else:
        environ['QUERY_STRING'] = ''

    if event_body:
        environ['CONTENT_LENGTH'] = str(len(event_body))
    else:
        environ['CONTENT_LENGTH'] = ''

    try:
        Request(environ)
        return environ
    except Exception as wtf:
        logger.error(wtf, exc_info=True)
        return None


class Response(object):
    def __init__(self):
        self.status = None
        self.response_headers = None

    def start_response(self, status, response_headers, exc_info=None):
        self.status = int(status[:3])
        self.response_headers = dict(response_headers)


class FlaskLambda(Flask):
    def get_event(self):
        return self.event

    def __call__(self, event, context):
        self.event = event
        self.request_data = create_request_data(event)
        if 'httpMethod' not in event:
            # In this "context" `event` is `environ` and
            # `context` is `start_response`, meaning the request didn't
            # occur via API Gateway and Lambda
            return super(FlaskLambda, self).__call__(event, context)

        response = Response()

        encoded_body = next(self.wsgi_app(
            make_environ(event),
            response.start_response
        ))

        body = encoded_body.decode('utf-8')
        logger.debug(f'encoded_body={encoded_body}')
        logger.debug(f'body={body}')

        content_type = response.response_headers.get('Content-Type')
        content_family = content_type.split('/')[0]
        if content_family in binary_things or content_type == OCTET_STREAM:
            wrk = {
                'statusCode': response.status,
                'headers': response.response_headers,
                'body': body,
                'isBase64Encoded': True
            }

            the_answer = json.dumps(wrk)
            logger.info(f'the_answer={the_answer}')
            return wrk
        else:
            return {
                'statusCode': response.status,
                'headers': response.response_headers,
                'body': body
            }


if __name__ == '__main__':
    make_environ(sample)



welcome_html = '''<html>
    <head>
        <title>Lambda Tool v{version}
    </head>
    <body>
        <div>
            <h3><a
                target="giithoob"
                href="https://github.com/muckamuck/lambda-tool/blob/master/README.md">Lambda
                Tool v{version}</a>:
            </h3>
            <p>A tool to create and deploy Lambda Functions to AWS (for python things).</p>
            <ul>
                <li>Create a new Python AWS Lambda from included template. Either a simple lambda function or a Flask based microservice.</li>
                <li>Deploy AWS Lambda created with this tool. It generates a CloudFormation file and creates a stack from that template.</li>
                <li>Optionally, integrate the new lambda with an AWS API Gateway, i.e. make a Flask application</li>
                <li>Optionally, subscribe the new lambda to an SNS topic</li>
                <li>Optionally, trust an arbitrary AWS service like cognito or S3</li>
            </ul>
        </div>
        <div>
            Zevon version: 0.5.0
        </div>
    </body>
</html>'''



sample = {
    "resource": "/{proxy+}",
    "path": "/work",
    "httpMethod": "POST",
    "headers": {
        "Accept": "*/*",
        "CloudFront-Forwarded-Proto": "https",
        "CloudFront-Is-Desktop-Viewer": "true",
        "CloudFront-Is-Mobile-Viewer": "false",
        "CloudFront-Is-SmartTV-Viewer": "false",
        "CloudFront-Is-Tablet-Viewer": "false",
        "CloudFront-Viewer-Country": "US",
        "content-type": "application/json",
        "Host": "66ivosvo25.execute-api.us-east-1.amazonaws.com",
        "User-Agent": "curl/7.68.0",
        "Via": "2.0 de2ed3c94563fee614f35f9bc3f52d1c.cloudfront.net (CloudFront)",
        "X-Amz-Cf-Id": "8LodH__BWnXodDK3Ejh-P6I7ZFKJYemE4ietXtkBd2nk7G5rH76haQ==",
        "X-Amzn-Trace-Id": "Root=1-60e8c842-304495781b0ba1e30b532d82",
        "X-Forwarded-For": "35.168.174.90, 130.176.98.91",
        "X-Forwarded-Port": "443",
        "X-Forwarded-Proto": "https"
    },
    "multiValueHeaders": {
        "Accept": [
            "*/*"
        ],
        "CloudFront-Forwarded-Proto": [
            "https"
        ],
        "CloudFront-Is-Desktop-Viewer": [
            "true"
        ],
        "CloudFront-Is-Mobile-Viewer": [
            "false"
        ],
        "CloudFront-Is-SmartTV-Viewer": [
            "false"
        ],
        "CloudFront-Is-Tablet-Viewer": [
            "false"
        ],
        "CloudFront-Viewer-Country": [
            "US"
        ],
        "content-type": [
            "application/json"
        ],
        "Host": [
            "66ivosvo25.execute-api.us-east-1.amazonaws.com"
        ],
        "User-Agent": [
            "curl/7.68.0"
        ],
        "Via": [
            "2.0 de2ed3c94563fee614f35f9bc3f52d1c.cloudfront.net (CloudFront)"
        ],
        "X-Amz-Cf-Id": [
            "8LodH__BWnXodDK3Ejh-P6I7ZFKJYemE4ietXtkBd2nk7G5rH76haQ=="
        ],
        "X-Amzn-Trace-Id": [
            "Root=1-60e8c842-304495781b0ba1e30b532d82"
        ],
        "X-Forwarded-For": [
            "35.168.174.90, 130.176.98.91"
        ],
        "X-Forwarded-Port": [
            "443"
        ],
        "X-Forwarded-Proto": [
            "https"
        ]
    },
    "queryStringParameters": None,
    "multiValueQueryStringParameters": None,
    "pathParameters": {
        "proxy": "work"
    },
    "stageVariables": None,
    "requestContext": {
        "resourceId": "9qz2u9",
        "resourcePath": "/{proxy+}",
        "httpMethod": "POST",
        "extendedRequestId": "COQ6fFKnIAMFUhw=",
        "requestTime": "09/Jul/2021:22:05:54 +0000",
        "path": "/dev/work",
        "accountId": "018734038160",
        "protocol": "HTTP/1.1",
        "stage": "dev",
        "domainPrefix": "66ivosvo25",
        "requestTimeEpoch": 1625868354933,
        "requestId": "0372d9c2-9f79-4368-9547-a1d418404da9",
        "identity": {
            "cognitoIdentityPoolId": None,
            "accountId": None,
            "cognitoIdentityId": None,
            "caller": None,
            "sourceIp": "35.168.174.90",
            "principalOrgId": None,
            "accessKey": None,
            "cognitoAuthenticationType": None,
            "cognitoAuthenticationProvider": None,
            "userArn": None,
            "userAgent": "curl/7.68.0",
            "user": None
        },
        "domainName": "66ivosvo25.execute-api.us-east-1.amazonaws.com",
        "apiId": "66ivosvo25"
    },
    "body": "eyJ4IjogNDJ9",
    "isBase64Encoded": True
}



from setuptools import setup

setup(
    name='Zevon',
    version='0.5.0',
    packages=['zevon'],
    description='Zevon - turn a API Gateway / Lambda event to a Flask thing',
    author='Chuck Muckamuck (obviously a pseudonym)',
    author_email='Chuck.Muckamuck@gmail.com',
    install_requires=[
        'Flask>=2.1',
        'Werkzeug<2.2.3'
    ]
)
